
#from IPython.Debugger import Tracer; debug_here = Tracer
from scipy import *
from scipy.linalg import *
from pylab import *
from scipy.integrate import *

#This code is a modification from that in elect.py (see the first electrophoresis problem in HW6)
#More details of how th code work can also be found in there.

#This code solves for the motion of a chain with 
#1. fixed link lengths, 
#2. inertia is #negligible (it always is for microscopic systems in a solvent) 
#3 A drag on each monomer proportional to the velocity
#4 Various other forces can be added to the simulation, like giving it an elastic
#  bending rigity, or pulling on the ends.
#5 It can be used to study electrophoresis by adding a periodic
# force field representing an array of obstacles.

#Because the links are constant in length, this makes it useful for studying many
#problems where the chain is close to being inextensible. However the algorithm
# to enforce inextensible links is a bit more involved. It involves calculating the
# tension along the chain as a function of position. Once the tension is calculated
# it is easy to add it to the other forces in the problem to get the equation of motion.

#This class is used to calculate the tension on the links of the chain.

class Tension:
    def __init__(self, T, n, d):
        self.T = T
        self.delta = zeros((n-1,d),dtype=double)
        self.delta_f = zeros((n-1,d),dtype=double)
        self.u = zeros((n-1),dtype=double)
        self.l = zeros((n-1),dtype=double)
        self.rhs = zeros((n-1,d),dtype=double)
        self.diag = zeros((n-1),dtype=double)
        self.force = zeros((n,d),dtype=double)
        self.lengths = ones((n-1),dtype=double)
        self.lengths[0] = 2


    def tension(self, r, f):
        r[0,:]=0
        f[0,:] = 0
        self.delta = r[1:,:]-r[:-1,:]
        self.diag = -2*sum(self.delta*self.delta,axis=1)
        self.diag[0] *= 0.5
        self.delta_f = f[1:,:]-f[:-1,:]
        u_vect = self.delta[1:,:]*self.delta[:-1,:]
        self.u[1:] = sum(u_vect, axis = 1)
        self.l[:-1] = self.u[1:]
        self.rhs = -sum(self.delta_f*self.delta,axis=1) - c1*(-0.5*self.diag*self.lengths - StepLengthSq)

        return solve_banded((1, 1), array((self.u,self.diag,self.l)), self.rhs, overwrite_ab=0, overwrite_b=0, debug=0)

    def calcForce(self,r,f):
        t = self.tension(r,f)
        for i in range(d):
            self.force[1:-1,i] = t[1:]*self.delta[1:,i] - t[0:-1]*self.delta[0:-1,i] + f[1:-1,i]
        #self.force[0,:] = t[0]*self.delta[0,:] + f[0,:]
        self.force[0,:] = 0
        self.force[n-1,:] = -t[n-2]*self.delta[n-2,:] + f[n-1,:]
        return self.force


def StiffForces(r):
    r[0,:]=0
    f_stiff[2:n-3,:] =  StiffnessConst*(2*r[2:n-3,:]-r[4:n-1,:] - r[0:n-5,:])
    f_stiff[1,:] = StiffnessConst*(r[1,:] - r[3,:])
    f_stiff[0,:] = StiffnessConst*(r[0,:] - r[2,:])
    f_stiff[n-2,:] = StiffnessConst*(r[n-2,:] - r[n-4,:])
    f_stiff[n-1,:] = StiffnessConst*(r[n-1,:] - r[n-3,:])
    return f_stiff


def Curvature(r):
    curve[1:n-1,:] = -2*r[1:n-1,:]+r[2:n,:] + r[0:n-2,:]
    return curve

def AbsCurve(r):
    curve = Curvature(r)
    abscurve[:] = sqrt(sum(curve[:,:]**2,axis=1))
    return abscurve
    
 
#The force due to the kinesin. It acts tangent to the directin of a microtubule
def KinesinForces(r):
    r[0,:]=0
    f_kin[1:n-1,:] =  KinesinConst*(r[2:n,:]-r[0:n-2,:])
    return f_kin

def ForceWall(r):
    r[0,:]=0
    f_wall[:,0] = (1.0/(r[:,0] + 1)-1)**2 * signbit(r[:,0])
    return f_wall


def dx_dt(r_in,t):
    r = reshape(r_in,(n,d))
    #remove the hash in the next line to add a wall:
    f[:,:] = StiffForces(r) + KinesinForces(r)#   + ForceWall(r)
    f[:,:] += f_const
    f[0,:] = 0
    r[0,:] = 0
    return tens.calcForce(r,f).flatten()

def length_error(r):
    delta = r[1:,:] - r[:-1,:]
    dlt_sq = delta*delta
    return sum(sum(dlt_sq)) - (n-1)

def construct_chain1(n):
    r = zeros((1,d))
    r1 = zeros((1,d))
    for i in range(n-1):
        dlt = standard_normal((d))
        dlt = StepLength*dlt/linalg.norm(dlt)
        r1 += reshape(dlt,(1,d))
        
        r = append(r, r1, 0)
    return r

def construct_chain(n):
    r = zeros((1,d))
    r1 = zeros((1,d))
    r_try = zeros((1,d))
    i=0
    while i < n-1:
        while True:
            dlt = standard_normal((d))
            dlt = StepLength*dlt/linalg.norm(dlt)
            r_try += reshape(dlt,(1,d))
            if r_try[0,0] > -0.5:
               break
        r1 = r_try
        r = append(r, r1, 0)
        i += 1
    return r

seed(100)
StepLength = 1.0
StepLengthSq = StepLength**2
#dimension of space
d = 3
#number of beads on the chain
n = 48
#time step used in diff eq solver:
dt = 0.1
#Temperature is set to zero because the energy scale for these effects is small compared to other things
T = 0.0
f_const = zeros((d),dtype=double)
#Change f_const to add a force due to an externally imposed streaming velocity.
f_const[1] = 0.4
c1 = 1 #helps to maintain constant link lengths
#used in calculating elastic (stiff) forces
f_stiff = zeros((n,d),dtype=double)
#used to calculate the curvature along the microtubule
curve = zeros((n,d),dtype=double)
abscurve = zeros((n),dtype=double)
#The force due to the kinesin. It acts tangent to the directin of a microtubule
f_kin = zeros((n,d),dtype=double)
f_wall = zeros((n,d),dtype=double)
#elastic constant of the microtubule (arbitrary units)
StiffnessConst = 10.0
#How strong is the force acting on the microtubule due to the kinesin. This will be proportional
#to the speed of the kinesin.
KinesinConst = -1.0

tens = Tension(T, n, d)

#r holds an array of the coordinates of the chain
r = construct_chain(n)

#time array
t = arange(0,10*dt,dt)

f = zeros((n,d))

graphics= True
if graphics:
    from visual import *
    ball = []
    cyl = []
    for i in range(n-1):
        cyl.append(cylinder (pos=r[i], axis=r[i+1]-r[i], color=color.blue, radius = 0.2))

    sphere = sphere(pos=(0,0,0),radius=0.25, color=color.green)

#if graphics:

ion()

x_out = r.flatten()

x_out = reshape(x_out,(1,n*d))

display_interval = 2

time = 0;
#debug_here
#raise NotImplementedError('Debug')
for t_coarse in arange(0,1000,dt):
    if graphics:
        rate(10000)
    t = array([t_coarse,t_coarse+dt])
    #We don't need to use this next line since T=0
#    noise = sqrt(2*T/dt)*standard_normal((n,d))
    #The diff'lequations are integrated. All the work comes in calculating dx_dt
    x_out = odeint(dx_dt, x_out[-1] , t, full_output=False)
    time += 1
    if time%10 == 0 : 
        print "length error = ", length_error(reshape(x_out[-1],(n,d)))
    if time%10 == display_interval : 
        r_out = reshape(x_out[-1],(n,d))
        if  graphics:
            for i in range(n-1):
                cyl[i].pos = r_out[i]
                cyl[i].axis = r_out[i+1]-r_out[i]

        elif True:
            cla()
            a = AbsCurve(r_out)
            plot(range(0,n),a)
            print a[2]-a[1]
            b =  Curvature(r_out)
            #plot(range(0,n),b[:,0],range(0,n),b[:,1],range(0,n),b[:,2])
            #draw()
            print "1/curv[N/2] = ", 1/a[n/2], "R[N/2] = ", sqrt(sum(r_out[n/2,:]*r_out[n/2,:]))
            print "r_out", r_out[n/2,:]



show()
