
from scipy import *
from scipy.linalg import *
from pylab import *
import scipy.integrate as integrate

#This code solves for the motion of a chain with 
#1. fixed link lengths, 
#2. inertia is #negligible (it always is for microscopic systems in a solvent) 
#3 A drag on each monomer proportional to the velocity
#4 Various other forces can be added to the simulation, like giving it an elastic
#  bending rigity, or pulling on the ends.
#5 In another file, similar code is used to study electrophoresis by adding a periodic
# force field representing an array of obstacles.

#Because the links are constant in length, this makes it useful for studying many
#problems where the chain is close to being inextensible. However the algorithm
# to enforce inextensible links is a bit more involved. It involves calculating the
# tension along the chain as a function of position. Once the tension is calculated
# it is easy to add it to the other forces in the problem to get the equation of motion.

#This class is used to calculate the tension on the links of the chain.
class Tension:
    def __init__(self, T, n, d):
        self.T = T
        self.delta = zeros((n-1,d),dtype=double)
        self.delta_f = zeros((n-1,d),dtype=double)
        self.u = zeros((n-1),dtype=double)
        self.l = zeros((n-1),dtype=double)
        self.rhs = zeros((n-1,d),dtype=double)
        self.diag = zeros((n-1),dtype=double)
        self.force = zeros((n,d),dtype=double)


    def tension(self, r, f):
        self.delta = r[1:,:]-r[:-1,:]
        self.diag = -2*sum(self.delta*self.delta,axis=1)
        self.delta_f = f[1:,:]-f[:-1,:]
        u_vect = self.delta[1:,:]*self.delta[:-1,:]
        self.u[1:] = sum(u_vect, axis = 1)
        self.l[:-1] = self.u[1:]
        self.rhs = -sum(self.delta_f*self.delta,axis=1) - c1*(-0.5*self.diag - StepLengthSq)

        return solve_banded((1, 1), array((self.u,self.diag,self.l)), self.rhs, overwrite_ab=0, overwrite_b=0, debug=0)

    def calcForce(self,r,f):
        t = self.tension(r,f)
        for i in range(d):
            self.force[1:-1,i] = t[1:]*self.delta[1:,i] - t[0:-1]*self.delta[0:-1,i] + f[1:-1,i]
        self.force[0,:] = t[0]*self.delta[0,:] + f[0,:]
        self.force[n-1,:] = -t[n-2]*self.delta[n-2,:] + f[n-1,:]
        return self.force


def StiffForces(r):
    f_stiff[2:n-3,:] =  StiffnessConst*(2*r[2:n-3,:]-r[4:n-1,:] - r[0:n-5,:])
    f_stiff[1,:] = StiffnessConst*(r[1,:] - r[3,:])
    f_stiff[0,:] = StiffnessConst*(r[0,:] - r[2,:])
    f_stiff[n-2,:] = StiffnessConst*(r[n-2,:] - r[n-4,:])
    f_stiff[n-1,:] = StiffnessConst*(r[n-1,:] - r[n-3,:])
    return f_stiff

def dx_dt(r_in,t):
    r = reshape(r_in,(n,d))
    f[:,:] = noise[:,:] + StiffForces(r)
    f[0,0] += -f_ext
    f[n-1,0] += f_ext
    return tens.calcForce(r,f).flatten()

def length_error(r):
    delta = r[1:,:] - r[:-1,:]
    dlt_sq = delta*delta
    return sum(sum(dlt_sq)) - (n-1)

def construct_chain(n):
    r = zeros((1,d))
    r1 = zeros((1,d))
    for i in range(n-1):
        dlt = standard_normal((d))
        dlt = StepLength*dlt/linalg.norm(dlt)
        r1 += reshape(dlt,(1,d))
        
        r = append(r, r1, 0)
    return r

   

StepLength = 1.0
StepLengthSq = StepLength**2
#dimension of space
d = 3
#number of beads on the chain
n = 32
#time step used in diff eq solver:
dt = 0.1
#Temperature:
T = 1.0
#Force applied to both ends:
f_ext = 10.0
c1 = 1 #helps to maintain constant link lengths
f_stiff = zeros((n,d),dtype=double)
#The elastic constant of the DNA (in rescaled units)
StiffnessConst = 0.0

tens = Tension(T, n, d)

#r holds an array of the coordinates of the chain
r = construct_chain(n)

#time array
t = arange(0,10*dt,dt)

f = zeros((n,d))

graphics= True
if graphics:
    from visual import *
    ball = []
    cyl = []
    for i in range(n-1):
        cyl.append(cylinder (pos=r[i], axis=r[i+1]-r[i], color=color.red, radius = 0.2))

    floor = box (pos=(0,-4,-4), length=n, height=0.5, width=4, color=color.blue)

if graphics:
    ion()


x_out = r.flatten()

x_out = reshape(x_out,(1,n*d))

display_interval = 10

time = 0;
for t_coarse in arange(0,100,dt):
    if graphics:
        rate(10000)
    t = array([t_coarse,t_coarse+dt])
    #thermal noise is generated and used to push the chain:
    noise = sqrt(2*T/dt)*standard_normal((n,d))
    #The diff'lequations are integrated. All the work comes in calculating dx_dt
    x_out = integrate.odeint(dx_dt, x_out[-1] , t, full_output=False)
    time += 1
    if time%300 == 0 : 
        #check to make sure the links haven't changed their lengths
        print length_error(reshape(x_out[-1],(n,d)))
    if time%display_interval== 0: 
        if  graphics:
            r_out = reshape(x_out[-1],(n,d))
            for i in range(n-1):
                cyl[i].pos = r_out[i]
                cyl[i].axis = r_out[i+1]-r_out[i]



